<p class="logo">
  <a href="{{ pathto('index') }}">
    <img class="logo" src="{{ pathto('_static/bioinformatics-sidebar-logo.svg', 1) }}"
         title="https://orange3-bioinformatics.readthedocs.io/en/latest/"/>
  </a>
</p>


<h3>About</h3>
<p>Bioinformatics add-on for Orange data mining software package.</p>
