/* This CSS snippet makes that the navbar collapse on medium screen already */
@media screen and (max-width: 950px){
    .wy-nav-top{display:block}
    .wy-nav-side{left:-300px}
    .wy-nav-side.shift{width:85%;left:0}
    .wy-side-scroll{width:auto}
    .wy-side-nav-search{width:auto}
    .wy-menu.wy-menu-vertical{width:auto}
    .wy-nav-content-wrap{margin-left:0}
    .wy-nav-content-wrap
    .wy-nav-content{padding:1.618em}
    .wy-nav-content-wrap.shift{position:fixed;min-width:100%;left:85%;top:0;height:100%;overflow:hidden}
}